@@ -4,7 +4,7 @@
     This file is part of Corrade.
 
     Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-                2017, 2018 Vladimír Vondruš <mosra@centrum.cz>
+                2017, 2018, 2019 Vladimír Vondruš <mosra@centrum.cz>
 
     Permission is hereby granted, free of charge, to any person obtaining a
     copy of this software and associated documentation files (the "Software"),
@@ -29,14 +29,13 @@
  * @brief Class @ref Corrade::PluginManager::AbstractManager, macro @ref CORRADE_PLUGIN_VERSION, @ref CORRADE_PLUGIN_REGISTER()
  */
 
-#include <map>
-#include <string>
-#include <vector>
-
 #include "Corrade/Containers/EnumSet.h"
+#include "Corrade/Containers/Pointer.h"
 #include "Corrade/PluginManager/PluginManager.h"
 #include "Corrade/PluginManager/visibility.h"
-#include "Corrade/Utility/Resource.h"
+#include "Corrade/Utility/StlForwardString.h"
+#include "Corrade/Utility/StlForwardVector.h"
+#include "Corrade/Utility/Utility.h"
 
 #ifdef CORRADE_TARGET_WINDOWS
 /* I didn't find a better way to circumvent the need for including windows.h */
@@ -83,8 +82,9 @@ enum class LoadState: unsigned short {
     WrongInterfaceVersion = 1 << 2,
 
     /**
-     * The plugin doesn't have any metadata file or the metadata file contains
-     * errors. Returned by @ref AbstractManager::load().
+     * The plugin doesn't have any associated `*.conf` metadata file or the
+     * metadata file contains errors. Returned by
+     * @ref AbstractManager::loadState() and @ref AbstractManager::load().
      * @partialsupport Not available on platforms without
      *      @ref CORRADE_PLUGINMANAGER_NO_DYNAMIC_PLUGIN_SUPPORT "dynamic plugin support".
      */
@@ -164,7 +164,7 @@ enum class LoadState: unsigned short {
     #endif
 };
 
-/** @debugoperatorenum{Corrade::PluginManager::LoadState} */
+/** @debugoperatorenum{LoadState} */
 CORRADE_PLUGINMANAGER_EXPORT Utility::Debug& operator<<(Utility::Debug& debug, PluginManager::LoadState value);
 
 /**
@@ -186,8 +186,15 @@ typedef Containers::EnumSet<LoadState> LoadStates;
 
 CORRADE_ENUMSET_OPERATORS(LoadStates)
 
+/** @debugoperatorenum{LoadStates} */
+CORRADE_PLUGINMANAGER_EXPORT Utility::Debug& operator<<(Utility::Debug& debug, PluginManager::LoadStates value);
+
+namespace Implementation {
+    struct StaticPlugin;
+}
+
 /**
-@brief Non-templated base for plugin managers
+@brief Base for plugin managers
 
 See @ref Manager and @ref plugin-management for more information.
  */
@@ -198,11 +205,6 @@ class CORRADE_PLUGINMANAGER_EXPORT AbstractManager {
         /** @brief Plugin version */
         static const int Version;
 
-        #ifndef DOXYGEN_GENERATING_OUTPUT
-        typedef void* (*Instancer)(AbstractManager&, const std::string&);
-        static void importStaticPlugin(std::string plugin, int _version, std::string interface, Instancer instancer, void(*initializer)(), void(*finalizer)());
-        #endif
-
         /** @brief Copying is not allowed */
         AbstractManager(const AbstractManager&) = delete;
 
@@ -369,53 +371,37 @@ class CORRADE_PLUGINMANAGER_EXPORT AbstractManager {
            faster than public virtual destructor */
         ~AbstractManager();
 
+    #ifdef DOXYGEN_GENERATING_OUTPUT
+    private:
+    #else
+    public:
+    #endif
+        struct CORRADE_PLUGINMANAGER_LOCAL Plugin;
+        typedef void* (*Instancer)(AbstractManager&, const std::string&);
+        static void importStaticPlugin(int version, Implementation::StaticPlugin& plugin);
+        static void ejectStaticPlugin(int version, Implementation::StaticPlugin& plugin);
+
     #ifdef DOXYGEN_GENERATING_OUTPUT
     private:
     #else
     protected:
     #endif
-        struct CORRADE_PLUGINMANAGER_LOCAL StaticPlugin;
-        struct CORRADE_PLUGINMANAGER_LOCAL Plugin;
-        struct CORRADE_PLUGINMANAGER_LOCAL GlobalPluginStorage;
-
         #ifndef CORRADE_PLUGINMANAGER_NO_DYNAMIC_PLUGIN_SUPPORT
         explicit AbstractManager(std::string pluginInterface, const std::vector<std::string>& pluginSearchPaths, std::string pluginDirectory);
         #else
         explicit AbstractManager(std::string pluginInterface);
         #endif
 
-        #ifndef CORRADE_PLUGINMANAGER_NO_DYNAMIC_PLUGIN_SUPPORT
-        std::string _pluginDirectory;
-        #endif
-
-        /* Initialize global plugin map. On first run it creates the instance
-           and fills it with entries from staticPlugins(). The reference is
-           then in constructor stored in _plugins variable to avoid at least
-           some issues with duplicated static variables on static builds. */
-        static CORRADE_PLUGINMANAGER_LOCAL GlobalPluginStorage& initializeGlobalPluginStorage();
-
-        GlobalPluginStorage& _plugins;
-
-        void* instantiateInternal(const std::string& plugin);
-        void* loadAndInstantiateInternal(const std::string& plugin);
+        Containers::Pointer<AbstractPlugin> instantiateInternal(const std::string& plugin);
+        Containers::Pointer<AbstractPlugin> loadAndInstantiateInternal(const std::string& plugin);
 
     private:
-        /* Temporary storage of all information needed to import static plugins.
-           They are imported to plugins() map on first call to plugins(),
-           because at that time it is safe to assume that all static resources
-           (plugin configuration files) are already registered. After that, the
-           storage is deleted and set to `nullptr` to indicate that static
-           plugins have been already processed.
-
-           The vector is accessible via function, not directly, because we don't
-           know initialization order of static members and thus the vector could
-           be uninitalized when accessed from CORRADE_PLUGIN_REGISTER(). */
-        CORRADE_PLUGINMANAGER_LOCAL static std::vector<StaticPlugin*>*& staticPlugins();
+        struct State;
 
         CORRADE_PLUGINMANAGER_LOCAL void registerDynamicPlugin(const std::string& name, Plugin* plugin);
 
         CORRADE_PLUGINMANAGER_LOCAL void registerInstance(const std::string& plugin, AbstractPlugin& instance, const PluginMetadata*& metadata);
-        CORRADE_PLUGINMANAGER_LOCAL void unregisterInstance(const std::string& plugin, AbstractPlugin& instance);
+        CORRADE_PLUGINMANAGER_LOCAL void reregisterInstance(const std::string& plugin, AbstractPlugin& oldInstance, AbstractPlugin* newInstance);
 
         #ifndef CORRADE_PLUGINMANAGER_NO_DYNAMIC_PLUGIN_SUPPORT
         CORRADE_PLUGINMANAGER_LOCAL LoadState loadInternal(Plugin& plugin);
@@ -425,51 +411,95 @@ class CORRADE_PLUGINMANAGER_EXPORT AbstractManager {
         CORRADE_PLUGINMANAGER_LOCAL LoadState unloadRecursiveInternal(Plugin& plugin);
         #endif
 
-        std::string _pluginInterface;
-        std::map<std::string, Plugin&> _aliases;
-        std::map<std::string, std::vector<AbstractPlugin*>> _instances;
+        Containers::Pointer<State> _state;
 };
 
+namespace Implementation {
+
+struct StaticPlugin {
+    /* Assuming both plugin and interface are static strings produced by the
+       CORRADE_PLUGIN_REGISTER() macro, so there's no need to make an allocated
+       copy of them, just a direct reference */
+    const char* plugin;
+    const char* interface;
+    AbstractManager::Instancer instancer;
+    void(*initializer)();
+    void(*finalizer)();
+    /* This field shouldn't be written to by anything else than
+       importStaticPlugin() / ejectStaticPlugin(). It's zero-initilized by
+       default and those use it to avoid inserting a single item to the linked
+       list more than once. */
+    StaticPlugin* next;
+};
+
+}
+
 /** @hideinitializer
-@brief Import static plugin
+@brief Import a static plugin
 @param name      Static plugin name (the same as defined with
     @ref CORRADE_PLUGIN_REGISTER())
 
-If static plugins are compiled into dynamic library or directly into the
-executable, they should be automatically loaded at startup thanks to
-@ref CORRADE_AUTOMATIC_INITIALIZER() and @ref CORRADE_AUTOMATIC_FINALIZER()
-macros.
-
-If static plugins are compiled into static library, they are not automatically
-loaded at startup, so you need to load them explicitly by calling
-@ref CORRADE_PLUGIN_IMPORT() at the beginning of `main()` function. You can
-also wrap these macro calls into another function (which will then be compiled
-into dynamic library or main executable) and use @ref CORRADE_AUTOMATIC_INITIALIZER()
-macro for automatic call:
+If you link static plugins to your executable, they can't automatically
+register themselves at startup to be known to
+@ref Corrade::PluginManager::Manager "PluginManager::Manager", and you need to
+load them explicitly by calling @ref CORRADE_PLUGIN_IMPORT() at the beginning
+of the @cpp main() @ce function. You can also wrap these macro calls in another
+function (which will then be compiled into a dynamic library or the main
+executable) and use the @ref CORRADE_AUTOMATIC_INITIALIZER() macro for an
+automatic call:
 
 @snippet PluginManager.cpp CORRADE_PLUGIN_IMPORT
 
 @attention This macro should be called outside of any namespace. If you are
     running into linker errors with `pluginImporter_*`, this could be the
-    problem. See @ref CORRADE_RESOURCE_INITIALIZE() documentation for more
+    reason. See @ref CORRADE_RESOURCE_INITIALIZE() documentation for more
     information.
- */
+
+Functions called by this macro don't do any dynamic allocation or other
+operations that could fail, so it's safe to call it even in restricted phases
+of application exection. It's also safe to call this macro more than once.
+
+@see @ref CORRADE_PLUGIN_EJECT()
+*/
+/* This "bundles" CORRADE_RESOURCE_INITIALIZE() in itself. Keep in sync. */
 #define CORRADE_PLUGIN_IMPORT(name)                                         \
     extern int pluginImporter_##name();                                     \
+    extern int resourceInitializer_##name();                                \
     pluginImporter_##name();                                                \
-    CORRADE_RESOURCE_INITIALIZE(name)
+    resourceInitializer_##name();
+
+/** @hideinitializer
+@brief Eject a previously imported static plugin
+@param name      Static plugin name (the same as defined with
+    @ref CORRADE_PLUGIN_REGISTER())
+
+Deregisters a plugin previously registered using @ref CORRADE_PLUGIN_IMPORT().
+
+@attention This macro should be called outside of any namespace. See the
+    @ref CORRADE_RESOURCE_INITIALIZE() macro for more information.
+
+Functions called by this macro don't do any dynamic allocation or other
+operations that could fail, so it's safe to call it even in restricted phases
+of application exection. It's also safe to call this macro more than once.
+*/
+/* This "bundles" CORRADE_RESOURCE_FINALIZE() in itself. Keep in sync. */
+#define CORRADE_PLUGIN_EJECT(name)                                          \
+    extern int pluginEjector_##name();                                      \
+    extern int resourceFinalizer_##name();                                  \
+    pluginEjector_##name();                                                 \
+    resourceFinalizer_##name();
 
 /** @brief Plugin version */
-#define CORRADE_PLUGIN_VERSION 5
+#define CORRADE_PLUGIN_VERSION 6
 
-/**
+/** @hideinitializer
 @brief Register static or dynamic lugin
 @param name          Name of static plugin (equivalent of dynamic plugin
      filename)
 @param className     Plugin class name
-@param interface     Interface name (the same as is defined with
-     CORRADE_PLUGIN_INTERFACE() in plugin base class)
-@hideinitializer
+@param interface     Interface name (the same as is defined by the
+    @ref Corrade::PluginManager::AbstractPlugin::pluginInterface() "pluginInterface()"
+    member of given plugin class)
 
 If the plugin is built as **static** (using CMake command
 @ref corrade-cmake-add-static-plugin "corrade_add_static_plugin()"), registers
@@ -493,15 +523,30 @@ See @ref plugin-management for more information about plugin compilation.
 
 @attention This macro should be called outside of any namespace. If you are
     running into linker errors with `pluginImporter_`, this could be the
-    problem.
+    reason.
 */
 #ifdef CORRADE_STATIC_PLUGIN
-#define CORRADE_PLUGIN_REGISTER(name, className, interface)                 \
-    inline void* pluginInstancer_##name(Corrade::PluginManager::AbstractManager& manager, const std::string& plugin) \
-        { return new className(manager, plugin); }                          \
+#define CORRADE_PLUGIN_REGISTER(name, className, interface_)                \
+    namespace {                                                             \
+        Corrade::PluginManager::Implementation::StaticPlugin staticPlugin_##name; \
+    }                                                                       \
     int pluginImporter_##name();                                            \
     int pluginImporter_##name() {                                           \
-        Corrade::PluginManager::AbstractManager::importStaticPlugin(#name, CORRADE_PLUGIN_VERSION, interface, pluginInstancer_##name, className::initialize, className::finalize); return 1; \
+        staticPlugin_##name.plugin = #name;                                 \
+        staticPlugin_##name.interface = interface_;                         \
+        staticPlugin_##name.instancer =                                     \
+            [](Corrade::PluginManager::AbstractManager& manager, const std::string& plugin) -> void* { \
+                return new className(manager, plugin);                      \
+            };                                                              \
+        staticPlugin_##name.initializer = className::initialize;            \
+        staticPlugin_##name.finalizer = className::finalize;                \
+        Corrade::PluginManager::AbstractManager::importStaticPlugin(CORRADE_PLUGIN_VERSION, staticPlugin_##name); \
+        return 1;                                                           \
+    }                                                                       \
+    int pluginEjector_##name();                                             \
+    int pluginEjector_##name() {                                            \
+        Corrade::PluginManager::AbstractManager::ejectStaticPlugin(CORRADE_PLUGIN_VERSION, staticPlugin_##name); \
+        return 1;                                                           \
     }
 #elif defined(CORRADE_DYNAMIC_PLUGIN)
 #define CORRADE_PLUGIN_REGISTER(name, className, interface)                 \
