@@ -36,6 +36,10 @@
 #include "Corrade/Utility/ConfigurationValue.h"
 #include "Corrade/Utility/visibility.h"
 
+#ifdef CORRADE_BUILD_DEPRECATED
+#include "Corrade/Utility/Macros.h"
+#endif
+
 namespace Corrade { namespace Utility {
 
 /**
@@ -133,9 +137,44 @@ Boolean options would cause parsing ambiguity so they are not allowed, but you
 can work around the limitation like this, for example:
 
 @snippet Utility.cpp Arguments-delegating-bool
+
+With @ref Flag::IgnoreUnknownOptions it's also possible for multiple subsystems
+to share just a subset of the same prefixed options, ignoring the unknown ones.
+However in order to have a good user experience, the first instance should
+always understand all options to be able to provide full help text and properly
+react to unknown options.
+
+@snippet Utility.cpp Arguments-delegating-ignore-unknown
 */
 class CORRADE_UTILITY_EXPORT Arguments {
     public:
+        /**
+         * @brief Flag
+         *
+         * @see @ref Flags, @ref Arguments(Flags),
+         *      @ref Arguments(const std::string&, Flags)
+         */
+        enum class Flag: std::uint8_t {
+            /**
+             * For prefixed arguments (constructed with
+             * @ref Arguments(const std::string&, Flags)) this makes
+             * @ref parse() ignore unknown options. See
+             * @ref Utility-Arguments-delegating for a complete overview about
+             * delegating options and usage of this flag.
+             *
+             * It's not allowed to use this flag on unprefixed arguments.
+             */
+            IgnoreUnknownOptions = 1 << 0
+        };
+
+        /**
+         * @brief Flags
+         *
+         * @see @ref Arguments(Flags),
+         *      @ref Arguments(const std::string&, Flags)
+         */
+        typedef Containers::EnumSet<Flag> Flags;
+
         /**
          * @brief Environment values
          *
@@ -150,7 +189,8 @@ class CORRADE_UTILITY_EXPORT Arguments {
          */
         static std::vector<std::string> environment();
 
-        explicit Arguments();
+        /** @brief Constructor */
+        explicit Arguments(Flags flags = {});
 
         /**
          * @brief Construct prefixed arguments
@@ -160,10 +200,10 @@ class CORRADE_UTILITY_EXPORT Arguments {
          * you want to handle them in separate steps. Prefixed version can have
          * only named arguments and long options.
          *
-         * See class documentation for example.
+         * See class documentation for an example.
          * @see @ref addSkippedPrefix()
          */
-        explicit Arguments(const std::string& prefix);
+        explicit Arguments(const std::string& prefix, Flags flags = {});
 
         /** @brief Copying is not allowed */
         Arguments(const Arguments&) = delete;
@@ -190,7 +230,7 @@ class CORRADE_UTILITY_EXPORT Arguments {
         /**
          * @brief Argument prefix
          *
-         * If the class was instantiated with @ref Arguments(const std::string&),
+         * If the class was instantiated with @ref Arguments(const std::string&, Flags),
          * returns the specified prefix. Otherwise returns empty string.
          */
         std::string prefix() const;
@@ -201,7 +241,7 @@ class CORRADE_UTILITY_EXPORT Arguments {
          * Returns @cpp true @ce if @ref parse() was successfully called,
          * @cpp false @ce otherwise.
          */
-        bool isParsed() const { return _isParsed; }
+        bool isParsed() const;
 
         /**
          * @brief Add mandatory argument
@@ -212,7 +252,7 @@ class CORRADE_UTILITY_EXPORT Arguments {
          *
          * @code{.shell-session}
          * Usage:
-         *   ./app argument
+         *   ./app [--] argument
          *
          * Arguments:
          *   argument          help text
@@ -224,6 +264,8 @@ class CORRADE_UTILITY_EXPORT Arguments {
          *
          * Only non-boolean options are allowed in the prefixed version, use
          * @ref addOption() instead.
+         *
+         * @see @ref addFinalOptionalArgument()
          */
         Arguments& addArgument(std::string key);
 
@@ -317,6 +359,8 @@ class CORRADE_UTILITY_EXPORT Arguments {
          *   --option          help text
          *                     (default: defaultValue)
          * @endcode
+         *
+         * @see @ref addNamedArgument(), @ref addFinalOptionalArgument()
          */
         Arguments& addOption(std::string key, std::string defaultValue = std::string()) {
             return addOption('\0', std::move(key), std::move(defaultValue));
@@ -325,10 +369,10 @@ class CORRADE_UTILITY_EXPORT Arguments {
         /**
          * @brief Add boolean option with both short and long key alternative
          *
-         * If the option is present, the option has `true` value, otherwise it
-         * has `false` value. Unlike above functions, the usage text does not
-         * display option value and you need to set help text with
-         * @ref setHelp() to make it appear in option list:
+         * If the option is present, the option has a @cpp true @ce value,
+         * otherwise it has a @cpp false @ce value. Unlike above functions, the
+         * usage text does not display the option value and you need to set a
+         * help text with @ref setHelp() to make it appear in option list:
          *
          * @code{.shell-session}
          * Usage:
@@ -368,6 +412,28 @@ class CORRADE_UTILITY_EXPORT Arguments {
             return addBooleanOption('\0', std::move(key));
         }
 
+        /**
+         * @brief Add final optional argument
+         *
+         * Always parsed as the last after all other unnamed arguments.
+         * Compared to arguments added with @ref addArgument() this one doesn't
+         * need to be present; compared to options added with @ref addOption()
+         * it doesn't need to be specified together with option name. There can
+         * be only one final optional argument. After calling
+         * @cpp addFinalOptionalArgument("argument") @ce the argument will be
+         * displayed in help text like the following:
+         *
+         * @code{.shell-session}
+         * Usage:
+         *   ./app [--] [argument]
+         *
+         * Arguments:
+         *   argument          help text
+         *                     (default: defaultValue)
+         * @endcode
+         */
+        Arguments& addFinalOptionalArgument(std::string key, std::string defaultValue = std::string());
+
         /**
          * @brief Skip given prefix
          *
@@ -437,7 +503,14 @@ class CORRADE_UTILITY_EXPORT Arguments {
          * Help text can be set only in the unprefixed version.
          * @see @ref setCommand()
          */
-        Arguments& setHelp(std::string help);
+        Arguments& setGlobalHelp(std::string help);
+
+        #ifdef CORRADE_BUILD_DEPRECATED
+        /** @brief @copybrief setGlobalHelp()
+         * @deprecated Use @ref setGlobalHelp() instead.
+         */
+        CORRADE_DEPRECATED("use setGlobalHelp() instead") Arguments& setHelp(std::string help);
+        #endif
 
         /**
          * @brief Set help text for given key
@@ -547,6 +620,10 @@ class CORRADE_UTILITY_EXPORT Arguments {
         bool isSet(const std::string& key) const;
 
     private:
+        enum class InternalFlag: std::uint8_t;
+        typedef Containers::EnumSet<InternalFlag> InternalFlags;
+        CORRADE_ENUMSET_FRIEND_OPERATORS(InternalFlags)
+
         struct CORRADE_UTILITY_LOCAL Entry;
 
         bool CORRADE_UTILITY_LOCAL skippedPrefix(const std::string& key) const;
@@ -561,7 +638,9 @@ class CORRADE_UTILITY_EXPORT Arguments {
 
         std::string valueInternal(const std::string& key) const;
 
-        bool _isParsed{false};
+        InternalFlags _flags;
+        /* not std::size_t so it fits into the padding after flags */
+        std::uint32_t _finalOptionalArgument{};
         std::string _prefix;
         std::string _command;
         std::string _help;
